import sys
import os
import logging

from ..headers.AnnotationHeader import AnnotationHeader
from ..headers.DataFilenameHeader import DataFilenameHeader
from ..headers.ImageSegmentIdentificationHeader import ImageSegmentIdentificationHeader
from ..headers.PrimaryHeader import PrimaryHeader
from .AbstractFile import AbstractFile

##
# @brief Lrit File class.
# @details Contains headers and data for an LRIT file. A given LRIT file can
# contain a variable number and types of headers, and this is taken into
# account.
# @warning This class does not correctly handle EMWIN data, which is a common
# format.
# @author Miguel Dominguez
# @date 2016
class LritFile(AbstractFile):

    ##
    # @fn __init__
    # @brief Constructor.
    # @details Initializes the logger and an empty data list
    def __init__(self):
        self.logger = logging.getLogger(__name__)
        self.flush()

    ##
    # @fn flush
    # @brief Clears out all data and begins fresh.
    def flush(self):
        self.dataBuffer = []
        self.headerDict = dict()

    def filetype(self):
        return self.headerDict[PrimaryHeader.headerIndex()].fileTypeCode

    @property
    def filename(self):
        """Return filename from stored headers"""
        if DataFilenameHeader.headerIndex() in self.headerDict:
            return self.headerDict[DataFilenameHeader.headerIndex()].filename
        elif AnnotationHeader.headerIndex() in self.headerDict:
            return self.headerDict[AnnotationHeader.headerIndex()].annotation
        return None

    ##
    # @fn toFile
    # @brief Writes data to a file.
    # @warning Assumes the file is complete.
    def toFile(self, product_dir):
        filename = self.filename

        # make sure non-null
        if filename:
            if ImageSegmentIdentificationHeader.headerIndex() in self.headerDict and filename[0:6] == "OR_ABI":
                segment_num = self.headerDict[ImageSegmentIdentificationHeader.headerIndex()].segmentSequenceNumber
                filename, file_ext = os.path.splitext(filename)
                filename = "".join((filename, "_Seg", str(segment_num), file_ext))
            fileData = bytearray(self.dataBuffer)
            self.logger.info('Writing file: {0}'.format(filename))
            with open(os.path.join(product_dir,filename),'w') as fileToWrite:
                fileToWrite.write(fileData)
        else:
            self.logger.info("Could not write file.  No file name!")

    ##
    # @fn getRawData
    # @brief returns data minus header info
    def getRawData(self):
        return self.dataBuffer[self.headerDict[PrimaryHeader.headerIndex()].allHeadersLength:]

    ##
    # @fn __str__
    # @brief Returns formatted header information as a string
    # @param[out] str Header information as a string
    def __str__(self):
        output = ""
        for header in self.headerDict.values():
            output += str(header)
        return output
