import sys
import os
import logging, multiprocessing
import szip

from .LritFile import LritFile
from ..headers.HeaderException import HeaderException
from ..headers.InvalidHeaderException import InvalidHeaderException
from ..headers.SkippedPrimaryHeaderException import SkippedPrimaryHeaderException
from ..headers.RepeatedHeaderException import RepeatedHeaderException
from ..headers.HeaderFactory import HeaderFactory
from ..headers.PrimaryHeader import PrimaryHeader
from ..headers.ImageSegmentIdentificationHeader import ImageSegmentIdentificationHeader

##
# @brief Lrit File Builder class.
# @details This Builder pattern was used to separate the relatively complex
# file-building logic from the file object as a container
# @author Miguel Dominguez
# @date 2016
class LritFileBuilder(object):

    BITS_PER_BYTE = 8

    ##
    # @fn __init__
    # @brief Constructor.
    # @details Initializes the state to the beginning, before any file data
    # is received or processed
    def __init__(self):
        self.logger = logging.getLogger(__name__)
        self.stateFunc = self.__processPrimaryHeader
        self.file = LritFile()
        self.fileHeaderBytesCount = -1
        self.fileBitsCount = -1
        self.__stateStr = '__processPrimaryHeader'
        self.needMoreData = False
        self.hasProduct = False
        self.datacount = 0


    ##
    # @fn __processPrimaryHeader
    # @brief Processes data, but expects that it will receive a Primary Header.
    # @details A Primary Header should be the first part of a file that is
    # received.  If some other header is received, then an exception is raised.
    def __processPrimaryHeader(self):
        self.__stateStr = '__processPrimaryHeader'
        #will get header object, unless header object is invalid, leading to corruption!
        headerRecordLength = (self.file.dataBuffer[1] << self.BITS_PER_BYTE) | self.file.dataBuffer[2]
        if len(self.file.dataBuffer) >= headerRecordLength:
            (newHeader,_) = HeaderFactory.getHeader(self.file.dataBuffer)
                #Check that the first header is the primary header, otherwise corruption!
            if type(newHeader) is not PrimaryHeader:
                raise SkippedPrimaryHeaderException(newHeader.headerIndex())
            else:
                self.file.headerDict[newHeader.headerIndex()] = newHeader
                self.fileBitsCount = self.file.headerDict[PrimaryHeader.headerIndex()].fileLengthBits
                self.fileHeaderBytesCount = self.file.headerDict[PrimaryHeader.headerIndex()].allHeadersLength
                self.fileHeaderBytesCount -= self.file.headerDict[PrimaryHeader.headerIndex()].headerRecordLength
                self.stateFunc = self.__processSecondaryHeader
            self.needMoreData = False
        else:
            self.needMoreData = True
        self.hasProduct = False

    ##
    # @fn __processSecondaryHeader
    # @brief Processes data, but expects that it will receive a Secondary Header
    # @details After the Primary Header there will be a sequence of secondary
    # headers that can vary in type and number.  This will continue to process
    # headers until the fileHeaderBytesCount (gotten from the Primary Header) is
    # fulfilled. If an unknown header type or a repeated header type is found,
    # raise an exception.
    # @todo Flag unlikely header combinations (e.g. image headers in text output)
    def __processSecondaryHeader(self):
        self.__stateStr = '__processSecondaryHeader'
        # get a slice of the data buffer past all known headers;
        # prevents headers from being removed from file buffer which
        # needs to be written w/ headers
        bufferslice = self.file.dataBuffer[
            sum(h.headerRecordLength for h in self.file.headerDict.itervalues()):]
        # peek at header length to see if you have enough data
        headerRecordLength = (bufferslice[1] << self.BITS_PER_BYTE) | bufferslice[2]
        if headerRecordLength <= len(bufferslice):
            #Attempt to get a new header, unless header object is invalid, leading to corruption!
            (newHeader, _) = HeaderFactory.getHeader(bufferslice)
            self.fileHeaderBytesCount -= newHeader.headerRecordLength
            self.logger.debug("Beginning File Bits Count: {0}".format(self.fileBitsCount))

            #Check if the header already exists.  If so, indicates corruption!
            if newHeader.headerIndex() in self.file.headerDict:
                raise RepeatedHeaderException(newHeader.headerIndex())

            #Otherwise probably a valid header
            else:
                self.file.headerDict[newHeader.headerIndex()] = newHeader
            if self.fileHeaderBytesCount == 0:
                self.stateFunc = self.__processData
            self.needMoreData = False
        else:
            self.needMoreData = True
        self.hasProduct = False

    ##
    # @fn __processFinish
    # @brief A stop state, this should really never be called
    # @todo Maybe this state should empty the buffer just in case it is called
    # to prevent memory leaks.
    def __processFinish(self):
        self.__stateStr = '__processFinish'
        #Dead state, this instantiation of the class should no longer process data
        #or go back to the beginning state.  PPDU will instantiate a new object
        self.needMoreData = False
        self.hasProduct = True
        self.logger.info("FILE COMPLETE - filename=%s", self.file.filename)

    ##
    # @fn __processData
    # @brief Adds data to buffer until it meets the expected length
    # @details There is a case here based on the fact that file length is given
    # in bits in the header, rather than bytes.  There is nothing else in LRIT
    # that would suggest that this would ever be the case, but I'm not sure.
    # There is also a case that occurs when we have too much data in the buffer,
    # but again that should theoretically never happen.
    def __processData(self):
        self.__stateStr = '__processData'

        # get the data length by subtracting the header lengths
        datalen = self.datacount - \
            sum(h.headerRecordLength for h in self.file.headerDict.itervalues())

        #self.logger.debug(self.file)
        #Mostly keep letting data accumulate unabated until the end of the file
        remainingBitsCount = self.fileBitsCount - datalen * self.BITS_PER_BYTE
        self.logger.debug("Remaining Bits Count: {0}".format(remainingBitsCount))
        if remainingBitsCount == 0:
            self.stateFunc == self.__processFinish
        #File not along byte boundaries, mask the last byte
        #To be honest this is not done by the C++ implemenation, so this behavior
        #may or may not be correct (the spec never specifies!)
        elif remainingBitsCount < 0 and remainingBitsCount > -1 * self.BITS_PER_BYTE:
            self.file.dataBuffer[-1] = self.__maskLastByte(remainingBitsCount * -1, self.file.dataBuffer[-1])
            self.stateFunc == self.__processFinish
        elif remainingBitsCount <= -1 * self.BITS_PER_BYTE:
            #Somehow we miscounted the number of bits
            self.logger.warning("Bit miscount: %d bits remaining", remainingBitsCount)
        self.needMoreData = True
        self.hasProduct = False

    ##
    # @fn __maskLastByte
    # @brief Mask the last partial byte to only include valid data
    # @warning Untested, because it is not clear whether this case can actually
    # ever happen.
    # @param[in] remainingBits The number of leftover bits after all full bytes
    # are accounted for
    # @param[in] The partial byte that needs to be masked
    # @param[out] output The masked partial last byte
    def __maskLastByte(self, remainingBits, lastByte):
        mask = 0x80
        for shift in range(remainingBits):
            mask |= (mask >> shift)
        return ((mask & lastByte) >> (self.BITS_PER_BYTE - remainingBits))

    ##
    # @fn __debugStr
    # @brief Convenient string outlining the state of the File Builder.
    # @todo Can probably be removed
    def __debugStr(self):
        print("File Header Bytes Count: {0}".format(self.fileHeaderBytesCount))
        print("File Bits Count: {0}".format(self.fileBitsCount))
        print("State: {0}".format(self.__stateStr))


    ##
    # @fn __riceDecompress
    # @brief performs rice decompression on CPPDU data
    # @details Uses the szip python wrapper for libaec to decompress data in
    # CP_PDU packets for images compressed using Rice compression.  Decompression
    # must be performed piecewise on each CPPDU packet and then appended onto
    # the transport file.
    def __riceDecompress(self,data):
        rice_header = self.file.headerDict[131]
        image_struct_header = self.file.headerDict[1]

        flags = rice_header.flags
        bits_per_pxl = image_struct_header.bitsPerPixel
        pxls_per_blk = rice_header.pixelsPerBlock
        pxls_per_scl = image_struct_header.columnCount
        str_data = "".join(map(chr,data))
        decomp = szip.decompress(str_data, flags, bits_per_pxl, pxls_per_blk, pxls_per_scl)
        return decomp

    ##
    # @fn processBuffer
    # @brief Process new input data.
    # @details The other process functions are part of a state machine to
    # control how data is processed at each point.  This function runs in
    # a loop until either there is not enough data to continue or an output
    # product is available.  Each state function self-reports these values.
    # @param[in] data Raw unsigned data bytes to add to buffer
    def processBuffer(self, data):
        # Check if we got a rice decompression header
        # If so, need to decompress data before adding it to the buffer
        self.datacount += len(data)
        if 131 in self.file.headerDict:
            data = self.__riceDecompress(data)

        self.file.dataBuffer += data
        self.logger.debug('File Data Length: {0}'.format(len(self.file.dataBuffer)))
        try:
            while not self.needMoreData and not self.hasProduct:
                self.stateFunc()
        except HeaderException as badHeader:
            #There is some sort of corruption in the data
            self.file.flush()
            self.stateFunc = self.__processPrimaryHeader
            print(badHeader.message)
            #Reraise this to someone who can log it
            self.logger.warn("Error in file headers, data corrupted.")
            #raise
